include_guard(GLOBAL)

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk)
	file(MAKE_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}/sdk)
endif()
message("--    SDK_TYPE : $ENV{SDK_TYPE}")
message("-- SDK_INCLUDE : $ENV{SDK_INCLUDE}")
message("--     SDK_LIB : $ENV{SDK_LIB}")

option(ENABLE_LL_DRIVER "Build & Link with LL drivers" OFF)
option(ENABLE_LL_DRIVER_USB "Build & Link with LL drivers usb support" OFF)
option(ENABLE_SEGGER_RTT "Build & Link with SEGGER RTT support, J-Link required!" OFF)
set(TARGET_CPU_FLOAT_ABI soft CACHE STRING "Select float ABI: none, soft, softfp, hard")
set(TARGET_SPECS nano CACHE STRING "Select specs: none, nosys, nano, rdimon, rdimon-v2m")
set(TARGET_XTAL_VALUE auto CACHE STRING "Select XTAL Speed: auto, <number>")
set(TARGET_XTAL32_VALUE auto CACHE STRING "Select XTAL32 Speed: auto, <number>")

if ("${TARGET_CPU}" STREQUAL "")
	message(FATAL_ERROR "TARGET_CPU was not specified!")
endif()

string(TOLOWER "${TARGET_CPU}" TARGET_CPU)
set(TARGET_CPU_TYPE cortex-m0)
# set(TARGET_CPU_TYPE cortex-m0plus)
# set(TARGET_CPU_TYPE cortex-m3)
# set(TARGET_CPU_TYPE cortex-m4)
# set(TARGET_CPU_TYPE cortex-m7)
# set(TARGET_CPU_TYPE cortex-m33)
# set(TARGET_CPU_TYPE cortex-a7)

if ("${TARGET_CPU}" MATCHES "^hc32f([4][0-9][0-9]).([0-9a-z])$")
	set(TARGET_CPU_TYPE cortex-m4)
	set(_HC_CPU_PREFIX HC32F${CMAKE_MATCH_1})
	set(_HC_CPU_SERIES ${CMAKE_MATCH_2})
	string(TOUPPER ${_HC_CPU_PREFIX} _HC_CPU_PREFIX)
	set(TARGET_CPU_TYPE_COMMON ${_HC_CPU_PREFIX})
	set(TARGET_CPU_TYPE_DEF ${_HC_CPU_PREFIX}x${_HC_CPU_SERIES})
	set(TARGET_CONF_FILE hc32f4xx_conf)
else()
	message(FATAL_ERROR "Unsupported device [${TARGET_CPU}]!")
endif()

message("--      Device : " ${TARGET_CPU})
message("--         CPU : " ${TARGET_CPU_TYPE})


if(CLANG)
	add_compile_options(--target=arm-none-eabi -Wno-unused-command-line-argument)
	add_link_options(--target=arm-none-eabi -Wno-unused-command-line-argument)
endif()

add_compile_definitions(${TARGET_CPU_TYPE_COMMON} ${TARGET_CPU_TYPE_DEF})
add_compile_options(
	-mthumb
	-mcpu=${TARGET_CPU_TYPE}
	-Wfatal-errors
	-fno-exceptions
	-fno-unwind-tables
	-fno-stack-check
)

set(CMAKE_CXX_FLAGS	"${CMAKE_CXX_FLAGS} -fno-rtti -fno-exceptions -fuse-cxa-atexit")
add_link_options(
	-mthumb
	-mcpu=${TARGET_CPU_TYPE}
	-Wl,--print-memory-usage
	-Wl,--warn-once
	${EMBEDED_GCC_EXTRA_LD_FLAGS}
	-L"$ENV{SDK_LIB}"
)

message("--       Specs : " ${TARGET_SPECS})
if (CLANG)
	if ("${TARGET_SPECS}" STREQUAL "nano")
		# alias as pico
		add_link_options(-lcrt0-minimal -ldummyhost)
	elseif ("${TARGET_SPECS}" STREQUAL "nosys")
		add_link_options(-lcrt0 -ldummyhost)
	elseif ("${TARGET_SPECS}" STREQUAL "rdimon")
		# should invoke initialise_monitor_handles
		add_link_options(-lcrt0-semihost -lsemihost)
	elseif ("${TARGET_SPECS}" STREQUAL "rdimon-v2m")
		# should invoke initialise_monitor_handles
		add_link_options(-lcrt0-semihost -lsemihost)
	elseif ("${TARGET_SPECS}" STREQUAL "none")
	else()
		message("${_WARNING_PREFIX} : Unsupported Specs " ${TARGET_SPECS})
	endif()
else()
	if ("${TARGET_SPECS}" STREQUAL "nano")
		add_link_options(--specs=nano.specs)
		add_link_options(-lnosys --specs=nosys.specs)
	elseif ("${TARGET_SPECS}" STREQUAL "nosys")
		add_link_options(-lnosys --specs=nosys.specs)
	elseif ("${TARGET_SPECS}" STREQUAL "rdimon")
		# should invoke initialise_monitor_handles
		add_link_options(-lrdimon --specs=rdimon.specs)
	elseif ("${TARGET_SPECS}" STREQUAL "rdimon-v2m")
		# should invoke initialise_monitor_handles
		add_link_options(-lrdimon-v2m --specs=rdimon-v2m.specs)
	elseif ("${TARGET_SPECS}" STREQUAL "none")
	else()
		message("${_WARNING_PREFIX} : Unsupported Specs " ${TARGET_SPECS})
	endif()
endif()

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/inc)
	file(MAKE_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}/inc)
endif()

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk)
	file(MAKE_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}/sdk)
endif()

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/inc/${TARGET_CONF_FILE}.h)
	string(TOUPPER "${TARGET_CONF_FILE}" __CONF_FILE_DEF)
	file(GENERATE OUTPUT "${CMAKE_CURRENT_SOURCE_DIR}/inc/${TARGET_CONF_FILE}.h"
		CONTENT
"#pragma once
#define __${__CONF_FILE_DEF}_H__

#ifdef __cplusplus
extern \"C\" {
#endif

#define LL_ICG_ENABLE                               (DDL_ON)
#define LL_UTILITY_ENABLE                           (DDL_ON)
#define LL_PRINT_ENABLE                             (DDL_ON)

#define LL_ADC_ENABLE                               (DDL_ON)
#define LL_AES_ENABLE                               (DDL_ON)
#define LL_AOS_ENABLE                               (DDL_ON)
#define LL_CAN_ENABLE                               (DDL_ON)
#define LL_CLK_ENABLE                               (DDL_ON)
#define LL_CMP_ENABLE                               (DDL_ON)
#define LL_CRC_ENABLE                               (DDL_ON)
#define LL_DBGC_ENABLE                              (DDL_ON)
#define LL_DCU_ENABLE                               (DDL_ON)
#define LL_DMA_ENABLE                               (DDL_ON)
#define LL_EFM_ENABLE                               (DDL_ON)
#define LL_EMB_ENABLE                               (DDL_ON)
#define LL_EVENT_PORT_ENABLE                        (DDL_ON)
#define LL_FCG_ENABLE                               (DDL_ON)
#define LL_FCM_ENABLE                               (DDL_ON)
#define LL_GPIO_ENABLE                              (DDL_ON)
#define LL_HASH_ENABLE                              (DDL_ON)
#define LL_I2C_ENABLE                               (DDL_ON)
#define LL_I2S_ENABLE                               (DDL_ON)
#define LL_INTERRUPTS_ENABLE                        (DDL_ON)
#define LL_INTERRUPTS_SHARE_ENABLE                  (DDL_ON)
#define LL_KEYSCAN_ENABLE                           (DDL_ON)
#define LL_MPU_ENABLE                               (DDL_ON)
#define LL_OTS_ENABLE                               (DDL_ON)
#define LL_PWC_ENABLE                               (DDL_ON)
#define LL_QSPI_ENABLE                              (DDL_ON)
#define LL_RMU_ENABLE                               (DDL_ON)
#define LL_RTC_ENABLE                               (DDL_ON)
#define LL_SDIOC_ENABLE                             (DDL_ON)
#define LL_SPI_ENABLE                               (DDL_ON)
#define LL_SRAM_ENABLE                              (DDL_ON)
#define LL_SWDT_ENABLE                              (DDL_ON)
#define LL_TMR0_ENABLE                              (DDL_ON)
#define LL_TMR4_ENABLE                              (DDL_ON)
#define LL_TMR6_ENABLE                              (DDL_ON)
#define LL_TMRA_ENABLE                              (DDL_ON)
#define LL_TRNG_ENABLE                              (DDL_ON)
#define LL_USART_ENABLE                             (DDL_ON)
#define LL_USB_ENABLE                               (DDL_OFF)
#define LL_WDT_ENABLE                               (DDL_ON)

#ifdef __cplusplus
}
#endif
"
	)
endif()

if (ENABLE_LL_DRIVER_USB)
	if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/inc/usb_app_conf.h)
		file(GENERATE OUTPUT "${CMAKE_CURRENT_SOURCE_DIR}/inc/usb_app_conf.h"
			CONTENT
"#pragma once
#define __USB_APP_CONF_H__

/* C binding of definitions if building with C++ compiler */
#ifdef __cplusplus
extern \"C\" {
#endif

/*******************************************************************************
* Include files
******************************************************************************/

/* USB MODE CONFIGURATION */
#define USB_FS_MODE
#define USE_DEVICE_MODE

#ifndef USB_FS_MODE
#define USB_FS_MODE
#endif

#ifndef USE_DEVICE_MODE
#ifndef USE_HOST_MODE
#error  \"USE_DEVICE_MODE or USE_HOST_MODE should be defined\"
#endif
#endif

/* USB DEVICE ENDPOINT CONFIGURATION */
#define CDC_IN_EP               (0x84U)
#define CDC_OUT_EP              (0x05U)
#define CDC_CMD_EP              (0x83U)

#define HID_IN_EP               (0x81U)
#define HID_OUT_EP              (0x02U)

/* USB FIFO CONFIGURATION */
#define RX_FIFO_FS_SIZE         (128U)
#define TX0_FIFO_FS_SIZE        (64U)
#define TX1_FIFO_FS_SIZE        (64U)
#define TX2_FIFO_FS_SIZE        (0U)
#define TX3_FIFO_FS_SIZE        (32U)
#define TX4_FIFO_FS_SIZE        (32U)
#define TX5_FIFO_FS_SIZE        (0U)

#if ((RX_FIFO_FS_SIZE + TX0_FIFO_FS_SIZE + TX1_FIFO_FS_SIZE + TX2_FIFO_FS_SIZE + TX3_FIFO_FS_SIZE + TX4_FIFO_FS_SIZE + \
TX5_FIFO_FS_SIZE) > 320U)
#error  \"The USB max FIFO size is 320 x 4 Bytes!\"
#endif

/* FUNCTION CONFIGURATION */
#define DEV_MAX_CFG_NUM         (1U)
#define USBD_ITF_MAX_NUM        (2U)
#define USB_MAX_STR_DESC_SIZ    (128U)

#define VBUS_SENSING_ENABLED
//#define SELF_POWER

/* CONFIGURATION FOR HID */
#define HID_IN_PACKET           (64U)
#define HID_OUT_PACKET          (64U)

/* CONFIGURATION FOR CDC */
#define MAX_CDC_PACKET_SIZE     (64U)      /* IN & OUT Endpoint Packet size */
#define CDC_CMD_PACKET_SIZE     (8U)       /* Control Endpoint Packet size */

#define CDC_IN_FRAME_INTERVAL   (5U)       /* Number of frames between IN transfers */
#define APP_RX_DATA_SIZE        (2048U)    /* Total size of IN buffer*/

#ifdef __cplusplus
}
#endif
"
	)
	endif()
endif()

add_compile_options(
	-I${CMAKE_CURRENT_SOURCE_DIR}/inc
)

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk/${TARGET_CPU_TYPE_DEF}.ld)
	if (EXISTS $ENV{SDK_CMSIS_DEVICES_SRC}/gcc/linker/${TARGET_CPU_TYPE_DEF}.ld)
	file(COPY_FILE
		$ENV{SDK_CMSIS_DEVICES_SRC}/gcc/linker/${TARGET_CPU_TYPE_DEF}.ld
		${CMAKE_CURRENT_SOURCE_DIR}/sdk/${TARGET_CPU_TYPE_DEF}.ld
	)
	endif()
endif()
# copy system_xxxx.c
if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk/system_${TARGET_CPU_TYPE_COMMON}.c)
	file(COPY_FILE
		$ENV{SDK_CMSIS_DEVICES_SRC}/system_${TARGET_CPU_TYPE_COMMON}.c
		${CMAKE_CURRENT_SOURCE_DIR}/sdk/system_${TARGET_CPU_TYPE_COMMON}.c
	)
endif()

# copy startup_xxxx.s
if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk/startup_${TARGET_CPU_TYPE_COMMON}.s)
	file(COPY_FILE
		$ENV{SDK_CMSIS_DEVICES_SRC}/GCC/startup_${TARGET_CPU_TYPE_COMMON}.s
		${CMAKE_CURRENT_SOURCE_DIR}/sdk/startup_${TARGET_CPU_TYPE_COMMON}.s
	)
endif()

# find ld script
if (EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/SDK/${TARGET_CPU_TYPE_DEF}.ld)
	add_link_options(-Wl,-T,${CMAKE_CURRENT_SOURCE_DIR}/SDK/${TARGET_CPU_TYPE_DEF}.ld)
	message("--   LD Script : sdk/${TARGET_CPU_TYPE_DEF}.ld")
elseif(EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/SDK/${TARGET_CPU_TYPE_COMMON}.ld)
	add_link_options(-Wl,-T,${CMAKE_CURRENT_SOURCE_DIR}/SDK/${TARGET_CPU_TYPE_COMMON}.ld)
	message("--   LD Script : sdk/${TARGET_CPU_TYPE_COMMON}.ld")
else()
	set(_LINK_FILE_PREFIX ${TARGET_CPU})
	string(TOUPPER "${_LINK_FILE_PREFIX}" _LINK_FILE_PREFIX)
	string(REGEX REPLACE "^HC32(....).*" "HC32\\1xx" _LINK_FILE_PREFIX "${_LINK_FILE_PREFIX}")
	if(EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/SDK/${_LINK_FILE_PREFIX}.ld)
		add_link_options(-Wl,-T,${CMAKE_CURRENT_SOURCE_DIR}/SDK/${_LINK_FILE_PREFIX}.ld)
		message("--   LD Script : sdk/${_LINK_FILE_PREFIX}.ld")
	else()
		message("${_WARNING_PREFIX} : LD script not found!")
	endif()
endif()

include_directories(${CMAKE_CURRENT_SOURCE_DIR}/inc $ENV{SDK_INCLUDE})

if ("${TARGET_CPU_FLOAT_ABI}" MATCHES "soft|softfp|hard")
	add_compile_options(-mfloat-abi=${TARGET_CPU_FLOAT_ABI})
	add_link_options(-mfloat-abi=${TARGET_CPU_FLOAT_ABI})
	message("--   Float ABI : " ${TARGET_CPU_FLOAT_ABI})
elseif(NOT "${TARGET_CPU_FLOAT_ABI}" STREQUAL "none")
	message("${_WARNING_PREFIX} : Invalid Float ABI ${TARGET_CPU_FLOAT_ABI}")
endif()

if ("${TARGET_XTAL_VALUE}" STREQUAL "auto")
	message("--  XTAL_VALUE : auto")
elseif("${TARGET_XTAL_VALUE}" MATCHES "^[0-9]+$")
	add_compile_definitions(XTAL_VALUE=${TARGET_XTAL_VALUE})
	message("--  XTAL_VALUE : " ${TARGET_XTAL_VALUE})
else()
	message("${_WARNING_PREFIX} : Invalid XTAL_VALUE ${TARGET_XTAL_VALUE}")
endif()

if ("${TARGET_XTAL32_VALUE}" STREQUAL "auto")
	message("--XTAL32_VALUE : auto")
elseif("${TARGET_XTAL32_VALUE}" MATCHES "^[0-9]+$")
	add_compile_definitions(XTAL32_VALUE=${TARGET_XTAL32_VALUE})
	message("--XTAL32_VALUE : " ${TARGET_XTAL32_VALUE})
else()
	message("${_WARNING_PREFIX} : Invalid XTAL32_VALUE ${TARGET_XTAL32_VALUE}")
endif()

set(XHSC_SDK)

if (ENABLE_LL_DRIVER)
	message("--   LL Driver : enabled")
	file(GLOB_RECURSE LL_DRIVER_SOURCES
		LIST_DIRECTORIES false
		$ENV{SDK_DRIVER_SRC}/*_ll.c
		$ENV{SDK_DRIVER_SRC}/*_ll_*.c
	)
	list(TRANSFORM LL_DRIVER_SOURCES REPLACE "\\\\" "/")

	set(LL_DRIVER_USB_SOURCES ${LL_DRIVER_SOURCES})
	list(FILTER LL_DRIVER_USB_SOURCES INCLUDE REGEX .*_ll_usb.c)
	list(FILTER LL_DRIVER_SOURCES EXCLUDE REGEX .*_ll_usb.c)
	add_library(ll-driver STATIC
		${LL_DRIVER_SOURCES}
	)
	list(APPEND XHSC_SDK ll-driver)

	if (ENABLE_LL_DRIVER_USB)
		add_library(ll-driver-usb STATIC
			${LL_DRIVER_USB_SOURCES}
		)
		list(APPEND XHSC_SDK ll-driver-usb)
		message("--   LL Driver : USB-Support")
	endif()

	add_compile_definitions(USE_DDL_DRIVER=1)
else()
	message("--   LL Driver : disabled")
endif()

include(${CMAKE_CURRENT_LIST_DIR}/segger-rtt.cmake)

file(GLOB_RECURSE SOURCE_FILES
	LIST_DIRECTORIES false
	${SEGGER_RTT_SOURCES}
	sdk/*.cxx
	sdk/*.cpp
	sdk/*.cc
	sdk/*.c
	sdk/*.s
	src/*.cxx
	src/*.cpp
	src/*.cc
	src/*.c
	src/*.s
)
